
--[[
Open Issues:

- GUID for players the same in battlegrounds?
]]

AbsorbsMonitor = LibStub("AceAddon-3.0"):NewAddon("AbsorbsMonitor", "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0", "AceComm-3.0", "AceSerializer-3.0");


-- #################################################################################################


-- In passive mode, no combat log parsing will be done
-- Meant for classes that want to export their realtime stats, but
-- not monitor absorb effects themselves
AbsorbsMonitor.isPassive = true;

-- In silent mode, no addon communcation through AceComm will take place
AbsorbsMonitor.isSilent = true;

local playerGUID;
local playerClass;

-- Specifies the channel any AddOn message should be
-- sent to, nil if silent
-- Can also be used to get the last known party state
local curChatChannel = nil;

-- Always hold the timestamp from the last COMBAT_LOG_EVENT_UNFILTERED fired
local lastCombatLogEvent = 0.0;

-- Table of all active absorb effects indexed by GUID and then spellId
-- at spellId == 0 there is a direct entry with the total remaining value,
-- at spellId == 1 there is a direct entry with the total quality
-- [GUID] = { [spellId] = {spellId, priority, remaining_shield_value, max_value, quality} }
local activeEffectsBySpell = {};

-- Table of all active absorb effects indexed by GUID and then in the order in which
-- they will be used
local activeEffectsByPriority = {};

-- Table of current unit charges
-- A charge is a variant value a custom trigger can put on any unit with a limited lifetime
-- It is used for example for Divine Aegis to save the critical heal value and then apply it on
-- the aura gain. (see Divine Aegis implementation why this is important)
local activeCharges = {};

-- Table of all known absorb effects (gets populated after the definition of the callbacks)
-- priority:	effects above 2 are neither used in total value nor displayed (e.g. Anti-Magic Shell)
-- [spell_id] = {priority, create_func, hit_func}
-- AbsorbsMonitor.Effects
local AbsorbEffects;

-- Table of additional callbacks on combat log events for proc-based and other non-generic absorb effects.
-- AbsorbsMonitor.CustomTriggers
local CustomTriggers;

-- Table of all unit stats relevant to absorb effects like attack power and spell power
-- (mastery rating later on)
-- [guid] = { [Class], [AttackPower], [SpellPower] }
AbsorbsMonitor.UnitStats = {};

-- Table of all scaling factors to absorb effects like talents, items, set boni, buffs
-- If there is no mechanism in Cataclysm to obtain the correct absorb amount by any effect
-- this entries are meant to be distributed among a group, raid, etc
-- [GUID] = { [scaling_Name] = scaling_Value }
-- A numerical scaling_name above 10 should ONLY be used if it is the spellId of the affected effect,
-- since it is sometimes used as a very quick way to check for a unit's class
-- Scaling factors that are only relevant to the local user like priest talents and do not
-- need to be distributed but are needed for calculation of the public one's are private ones,
-- found at index -1
AbsorbsMonitor.Scaling = {[-1] = {}};


-- #################################################################################################


local CommStatsCooldown = false;

local function ClearCommStatsCooldown()
	CommStatsCooldown = false;
end

local CommScalingCooldown = false;

local function ClearCommScalingCooldown()
	CommScalingCooldown = false;
end

local function DummyFunc()
end

local function DeepTableCopy(src)
	dest = {};
	
	for k, v in pairs(src) do
		if(type(k) == "table") then
			k = DeepTableCopy(k);
		end
		
		if(type(v) == "table") then
			v = DeepTableCopy(v);
		end
		
		dest[k] = v;
	end
	
	setmetatable(dest, getmetatable(src));
	
	return dest;
end

local function SortEffects(a, b)
	if(a[1] == b[1]) then
		return (a[3] < b[3]);
	else
		return (a[1] < b[1]);
	end
end


function AbsorbsMonitor:ApplyActiveEffect(sourceGUID, sourceName, destGUID, destName, spellId, ...)
	local destEntry = activeEffectsBySpell[destGUID];
	
	-- To help me find typos
	if(not AbsorbEffects[spellId][2]) then
		self:Print("create func nil on spellId "..spellId);
	end
		
	local value, quality = AbsorbEffects[spellId][2](sourceGUID, sourceName, destGUID, destName, spellId, destEntry, ...);
	
	if(value == nil) then return; end
	
	-- no entry yet for this unit
	if(not destEntry) then
		local effectEntry = {spellId, AbsorbEffects[spellId][1], value, value, 1.0};
		
		activeEffectsBySpell[destGUID] = {[0] = 0, [1] = 1.0, [spellId] = effectEntry};
		activeEffectsByPriority[destGUID] = {effectEntry};
		
		destEntry = activeEffectsBySpell[destGUID];
		
	-- not this specific effect yet
	elseif(not destEntry[spellId]) then	
		local effectEntry = {spellId, AbsorbEffects[spellId][1], value, value, 1.0};
				
		destEntry[spellId] = effectEntry;
		
		table.insert(activeEffectsByPriority[destGUID], effectEntry);
		table.sort(activeEffectsByPriority[destGUID], SortEffects);
		
	-- effect exists already
	else
		local prevAmount = destEntry[spellId][3];
	
		destEntry[spellId][3] = value;
		destEntry[spellId][4] = value;
		
		-- adjust value in case this is a visible absorb
		value = value - prevAmount;
	end
	
	if(quality < activeEffectsBySpell[destGUID][1]) then
		activeEffectsBySpell[destGUID][1] = quality;
	end
	
	if(AbsorbEffects[spellId][1] < 2) then
		destEntry[0] = destEntry[0] + value;
	
		self:Unit_UpdatedTotal(destGUID, destEntry[0], destEntry[1]);
	end
		
	--self:Print("created absorb effect "..spellId.." (priority: "..AbsorbEffects[spellId][1]..") on "..destGUID.." for "..value..", new total: "..activeEffectsBySpell[destGUID][0].." ("..#(activeEffectsByPriority[destGUID])..")");
end

function AbsorbsMonitor:RemoveActiveEffect(guid, spellId)
	local guidEntry = activeEffectsBySpell[guid];
	
	--self:Print("removing absorb effect "..spellId.." on "..guid.." with "..guidEntry[spellId][3].." left, total: "..guidEntry[0].." ("..#(activeEffectsByPriority[guid])..")");

	if(AbsorbEffects[spellId][1] < 2 and guidEntry[0]) then
		guidEntry[0] = guidEntry[0] - guidEntry[spellId][3];		
	end
	
	if(#(activeEffectsByPriority[guid]) == 1) then		
		activeEffectsBySpell[guid] = nil;
		activeEffectsByPriority[guid] = nil;
		
		self:Unit_Cleared(guid);
	else
		guidEntry[spellId] = nil;
		
		for k, v in pairs(activeEffectsByPriority[guid]) do
			if(v[1] == spellId) then
				table.remove(activeEffectsByPriority[guid], k);
				
				break;
			end
		end
		
		self:Unit_UpdatedTotal(guid, guidEntry[0], guidEntry[1]);
	end
end

function AbsorbsMonitor:HitUnit(guid, absorbedTotal, overkill, spellSchool)
	if(not activeEffectsBySpell[guid]) then return; end
	
	local guidEntry = activeEffectsBySpell[guid];
	
	local absorbedRemaining = absorbedTotal;
	
	local absorbed = 0;
	local keepEffect = false;
	local i = 1;
	local curEffect;
	local visibleAbsorb = false;	
	
	-- This loop lasts as long as there is still an absorb value that
	-- no effect could account for, but it will break once the list of
	-- available effects got used completely.
	while(absorbedRemaining > 0) do
		curEffect = activeEffectsByPriority[guid][i];
		
		-- Workaround for some really weird and non-rational bug
		if(curEffect == nil) then break; end
		
		if(curEffect[3] > 0) then
			-- Hit the abosrb effect
			absorbed, keepEffect = AbsorbEffects[curEffect[1]][3](guid, curEffect, absorbedRemaining, overkill, spellSchool);
			
			if(absorbed > 0) then
				-- If it should be visible (priority < 2), correct the total value
				if(curEffect[2] < 2) then
					guidEntry[0] = guidEntry[0] - absorbed;
				
					-- Shows us that at least one visible effect got hit
					visibleAbsorb = true;						
				end
			
				-- Reduce the value of this effect and the remaining absorb value
				-- to be accounted for
				curEffect[3] = curEffect[3] - absorbed;	
				absorbedRemaining = absorbedRemaining - absorbed;
			
				--self:Print("Hit effect #"..i..", absorbed "..absorbed.." (keep: "..tostring(keepEffect).."), remaining: "..absorbedRemaining.." - unit: "..curEffect[3].."/"..activeEffectsBySpell[guid][0]);
			end
			
			-- If the hit-function told us to remove the effect, do so
			-- Note that only AbsorbsMonitor:RemoveActiveEffect is allowed to
			-- remove it from the list, we just set the value to zero, so it gets
			-- ignored on any hit. This is do not come into any desync issues with
			-- the events, and to keep the proper clean-up code in one place
			if(not keepEffect) then				
				curEffect[3] = 0;					
			end
		end
			
		i = i + 1;
	end	
	
	-- there are two possibilities when things are going wrong
	-- 
	--	a)	we guessed an absorb value too high, in that case it will
	--		automatically be corrected when it breaks
	--
	--	b)	we guessed an absorb value too low, so we end up with an
	--		amount to absorb when all effects seem to be gone
	--		(absorbedRemaining > 0)
	--		Note that we cannot rely on SPELL_AURA_REMOVED to check this,
	--		since it may happen completely out of order, but it will
	--		clear this unit soon or did so already.
	--		we reduce the quality to zero, since any absorb now happening
	--		cannot be accounted for.
	--		Since we may have rounding errors from scanning the spellbook
	--		and calculating the value thereafter (does Blizzard round on
	--		EVERY step?!?), we accept a small threshold
	
	-- TODO: optimize that deep table access
	if(absorbedRemaining > self.db.profile.lowValueTolerance) then
		guidEntry[0] = guidEntry[0] - absorbedRemaining;
		guidEntry[1] = 0.0;
		visibleAbsorb = true;
	end
	
	if(visibleAbsorb) then
		self:Unit_UpdatedTotal(guid, guidEntry[0], guidEntry[1]);
	end	
end

-- Note that this implementation prolongs the lifetime of any charges if a new
-- one is added before the previous charges expire, it is therefore working like
-- a refreshable stack of buffs.
-- This is relevant since the Divine Aegis implementation cannot rely on this to
-- filter out those spells that do not proc it, since a non-proccing spell might be
-- followed by a proccing spell, extending the former's lifetime and therefore
-- counting it
function AbsorbsMonitor:AddCharge(guid, spellId, amount, lifetime)
	if(not activeCharges[guid]) then
		activeCharges[guid] = {[spellId] = {amount, lastCombatLogEvent + lifetime}}
		
	elseif(not activeCharges[guid][spellId]) then
		activeCharges[guid][spellId] = {amount, lastCombatLogEvent + lifetime};
	
	else
		local spellCharges = activeCharges[guid][spellId];
	
		if(lastCombatLogEvent > spellCharges[2]) then
			spellCharges[1] = amount;
		else
			spellCharges[1] = spellCharges[1] + amount;
		end
		
		spellCharges[2] = lastCombatLogEvent + lifetime;	
	end
end

function AbsorbsMonitor:ConsumeCharges(guid, spellId)
	if(activeCharges[guid] and activeCharges[guid][spellId]) then			
		local charge = activeCharges[guid][spellId][1];
		activeCharges[guid][spellId][1] = 0;	
	
		if(lastCombatLogEvent > activeCharges[guid][spellId][2]) then
			self:Print("expired: "..lastCombatLogEvent.." > "..activeCharges[guid][spellId][2]);
			return 0;
		else
			return charge;
		end
	end
	
	return 0;
end

function AbsorbsMonitor:AddCustomTrigger(target, event, func)
	local index = target.."_"..event;
	local oldTrigger = rawget(CustomTriggers, index);

	if(not oldTrigger) then
		CustomTriggers[index] = func;
		
	else
		local listIndex = index.."_list";
		local funcList = rawget(CustomTriggers, listIndex);
		
		-- There is already a list of callbacks, so we just add this one
		if(funcList) then
			for k, v in pairs(funcList) do
				if(v == func) then
					return;
				end
			end
		
			table.insert(funcList, func);
			
		-- We used a direct call so far, create a list and set up a handler
		else
			if(oldTrigger == func) then
				return;
			end
		
			funcList = {oldTrigger, func};			
			CustomTriggers[listIndex] = funcList;
			
			local handler = function(...)
				for k, v in pairs(funcList) do
					v(...);
				end
			end
			
			CustomTriggers[index] = handler;
		end		
	end
end

function AbsorbsMonitor:RemoveCustomTrigger(target, event, func)
	local index = target.."_"..event;
	
	local listIndex = index.."_list";
	local funcList = rawget(CustomTriggers, listIndex);
	
	-- We have a list of callbacks, reduce if possible
	if(funcList) then
		if(#funcList == 2) then
			CustomTriggers[index] = (funcList[1] == func) and funcList[2] or funcList[1];
			CustomTriggers[listIndex] = nil;
		else
			-- ATTENTION: We have to keep the table in place
			-- because the handler references this table
			
			local old_funcList = DeepTableCopy(funcList);
			wipe(funcList);
			
			for k, v in pairs(old_funcList) do
				if(v ~= func) then				
					table.insert(funcList, v);
				end
			end
		end
	
	-- It was a direct call anyway
	else
		CustomTriggers[index] = nil;
	end	
end

function AbsorbsMonitor:SetActive()
	if(not self.isPassive) then return; end

	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED", "OnCombatLogEvent");		
	
	self.isPassive = false;
end

function AbsorbsMonitor:SetPassive()
	if(self.isPassive) then return; end
	
	self:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
	self.isPassive = true;
end

function AbsorbsMonitor:SetVerbose()
	if(not self.isSilent) then return; end
	
	self:RegisterEvent("PARTY_MEMBERS_CHANGED", "OnPartyEvent");
	self:RegisterEvent("RAID_ROSTER_UPDATE", "OnPartyEvent");
	self.isSilent = false;
	
	self:OnPartyEvent();
end

function AbsorbsMonitor:SetSilent()
	if(self.isSilent) then return; end
	
	self:UnregisterEvent("PARTY_MEMBERS_CHANGED");
	self:UnregisterEvent("RAID_ROSTER_UPDATE");
	self.isSilent = true;
end

function AbsorbsMonitor:OnCombatLogEvent(_, timestamp, type, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags, ...)
	lastCombatLogEvent = timestamp;

	if(type == "SWING_MISSED") then
		local missType, amountMissed = select(1, ...);
		
		if(missType ~= "ABSORB") then return; end	
		
		self:HitUnit(destGUID, amountMissed, 0, PHYSICAL_SPELL_SCHOOL);
		
	elseif(type == "RANGE_MISSED" or type == "SPELL_MISSED" or type == "SPELL_PERIODIC_MISSED") then
		local spellId, spellName, spellSchool, missType, amountMissed = select(1, ...);
			
		if(missType ~= "ABSORB") then return; end
		
		self:HitUnit(destGUID, amountMissed, 0, spellSchool);
		
	elseif(type == "SWING_DAMAGE") then
		local amount, overkill, school, resisted, blocked, absorbed = select(1, ...);
			
		if(not absorbed) then return; end
		
		self:HitUnit(destGUID, absorbed, amount, PHYSICAL_SPELL_SCHOOL);
		
	elseif(type == "RANGE_DAMAGE" or type == "SPELL_DAMAGE" or type == "SPELL_PERIODIC_DAMAGE") then
		local spellId, spellName, spellSchool, amount, overkill, school, resisted, blocked, absorbed = select(1, ...);
		
		if(not absorbed) then return; end
		
		self:HitUnit(destGUID, absorbed, amount, spellSchool);
		
	elseif(type == "SPELL_HEAL") then
		local spellId, spellName, spellSchool, amount, overhealing, absorb, critical = select(1, ...);		
		
		CustomTriggers[sourceGUID.."_OnHeal"](sourceGUID, sourceName, destGUID, destName, spellId, amount, overhealing);
		
		if(critical) then
			CustomTriggers[sourceGUID.."_OnHealCrit"](sourceGUID, sourceName, destGUID, destName, spellId, amount, overhealing);
		end
	elseif(type == "SPELL_AURA_APPLIED") then
		local spellId = select(1, ...);
		
		if(AbsorbEffects[spellId]) then
			self:ApplyActiveEffect(sourceGUID, sourceName, destGUID, destName, spellId);
		end
		
		CustomTriggers[spellId.."_OnAuraApplied"](sourceGUID, sourceName, destGUID, destName, spellId);
		
	elseif(type == "SPELL_AURA_REFRESH") then
		local spellId = select(1, ...);
		
		if(AbsorbEffects[spellId]) then
			self:ApplyActiveEffect(sourceGUID, sourceName, destGUID, destName, spellId);
		end	
		
		CustomTriggers[spellId.."_OnAuraApplied"](sourceGUID, sourceName, destGUID, destName, spellId);
				
	elseif(type == "SPELL_AURA_REMOVED") then
		local spellId = select(1, ...);			
		
		if(AbsorbEffects[spellId]) then
			if(activeEffectsBySpell[destGUID] and activeEffectsBySpell[destGUID][spellId]) then
				self:RemoveActiveEffect(destGUID, spellId);
			end
		end
		
		CustomTriggers[spellId.."_OnAuraRemoved"](sourceGUID, sourceName, destGUID, destName, spellId);
		
	end
end

function AbsorbsMonitor:OnPartyEvent(...)
	-- Note that the order here is VERY important
	if(UnitInBattleground("player")) then	
		curChatChannel = "BATTLEGROUND";
	elseif(UnitInRaid("player")) then
		curChatChannel = "RAID";
	elseif(GetNumPartyMembers() > 0) then
		curChatChannel = "PARTY";
	else
		curChatChannel = nil;
	end	
end

-- Come on, this whole stuff is bullshit anyway
-- and will hopefully be removed in Cataclysm
function AbsorbsMonitor:OnZoneChangedNewArea()
	local profile = self.db.profile;
	
	if(not profile.zoneModifier_auto) then return; end

	if(UnitInBattleground("player")) then
		profile.zoneModifier = 1.17;
	elseif(IsActiveBattlefieldArena()) then
		profile.zoneModifier = 0.9;
	elseif(GetRealZoneText() == "Wintergrasp") then
		-- 1.3 * 0.9 = fail Blizzard
		profile.zoneModifier = 1.17;
	elseif(GetRealZoneText() == "Icecrown Citadel") then
		profile.zoneModifier = 1.3;
	else
		profile.zoneModifier = 1.3;
	end
end

function AbsorbsMonitor:OnScalingBroadcast()
	return self.Scaling[playerGUID];
end

function AbsorbsMonitor:OnUnitStatsReceived(prefix, text, distribution, target)
	if(not text) then return; end
	
	local success, guid, class, ap, sp = self:Deserialize(text);
	
	if(not(success and guid and class and ap and sp)) then return; end
	
	if(guid == playerGUID) then return; end
	
	if(not self.UnitStats[guid]) then
		self.UnitStats[guid] = {class, ap, sp};		
	else
		self.UnitStats[guid][2] = ap;
		self.UnitStats[guid][3] = sp;
	end
end

function AbsorbsMonitor:OnScalingReceived(prefix, text, distribution, target)
	if(not text) then return; end
	
	local success, guid, in_scaling = self:Deserialize(text);
	
	if(not(success and guid and in_scaling)) then return; end
	
	if(guid == playerGUID) then return; end
	
	self.Scaling[guid] = in_scaling;
end

function AbsorbsMonitor:OnInitialize()
	self:InitOptions();
	
	-- To actually see syntax errors in Effects.lua
	if(not self.Triggers) then return; end
	
	setmetatable(self.Triggers, {
		__index = function()
			return DummyFunc;
		end
	});
	
	AbsorbEffects = self.Effects;
	CustomTriggers = self.Triggers;
end

function AbsorbsMonitor:OnEnable()	
	_, playerClass = UnitClass("player");
	 playerGUID = UnitGUID("player");
	
	self.UnitStats[playerGUID] = {playerClass, UnitAttackPower("player"), GetSpellBonusHealing()};	
	self.Scaling[playerGUID] = {};	
	
	self:RegisterEvent("ZONE_CHANGED_NEW_AREA", "OnZoneChangedNewArea");
	self:OnZoneChangedNewArea();

	if(playerClass == "DEATHKNIGHT") then
		self:RegisterEvent("UNIT_ATTACK_POWER", "OnStatsChanged");
		
	elseif(playerClass == "DRUID") then
		self:RegisterEvent("UNIT_ATTACK_POWER", "OnStatsChanged");
		
	elseif(playerClass == "MAGE") then
		self:RegisterEvent("PLAYER_DAMAGE_DONE_MODS", "OnStatsChanged");
		
	elseif(playerClass == "PALADIN") then
		self:RegisterEvent("PLAYER_DAMAGE_DONE_MODS", "OnStatsChanged");
		
	elseif(playerClass == "PRIEST") then
		self:RegisterEvent("PLAYER_DAMAGE_DONE_MODS", "OnStatsChanged");
		
	elseif(playerClass == "WARLOCK") then
		self:RegisterEvent("PLAYER_DAMAGE_DONE_MODS", "OnStatsChanged");
	end
	
	self:EnableEffects();
	
	if(self['OnEnable_'..playerClass]) then
		self['OnEnable_'..playerClass](self);
	end
	
	if(self.OnLevelUp) then
		self:RegisterEvent("PLAYER_LEVEL_UP", "OnLevelUp");
	end
	
	if(self.OnTalentUpdate) then
		self:RegisterEvent("PLAYER_TALENT_UPDATE", "OnTalentUpdate");
	end
	
	if(self.OnGlyphUpdated) then
		self:RegisterEvent("GLYPH_UPDATED", "OnGlyphUpdated");
	end
	
	if(self.OnEquipmentChanged) then
		self:RegisterEvent("PLAYER_EQUIPMENT_CHANGED", "OnEquipmentChanged");
	end
	
	self:RegisterComm("Absorbs_UnitStats", "OnUnitStatsReceived");
	self:RegisterComm("Absorbs_Scaling	", "OnScalingReceived");	
	
	self:Print("Enabled");	
	
	if(not self.db.profile.passive) then
		self:SetActive();
	end
	
	if(not self.db.profile.silent) then
		self:SetVerbose();
	end
end

function AbsorbsMonitor:OnDisable()
	if(self['OnDisable_'..playerClass]) then
		self['OnDisable_'..playerClass]();
	end

	self:Print("Disabled");
end

function AbsorbsMonitor:OnStatsChanged(event, unit)
	if(curChatChannel) then
		self:ScheduleUniqueTimer("comm_stats", "Comm_UnitStats", CommStatsCooldown and 20 or 5);
	end
end

function AbsorbsMonitor:ScheduleScalingBroadcast()
	if(curChatChannel) then
		self:ScheduleUniqueTimer("comm_scaling", "Comm_Scaling", CommScalingCooldown and 20 or 5);
	end
end

-- An extension of AceTimer to schedule a one-shot timer that will not
-- be scheduled twice if scheduled again before it's fired.
-- Note that the callback itself has to make sure the timer gets marked
-- as done
-- I know this could also be done with a custom callback clearing it,
-- but then I have to mimic AceTimer's API concerning the callback
local activeTimers = {};

function AbsorbsMonitor:ScheduleUniqueTimer(id, callback, delay, arg)
	if(not activeTimers[id]) then
		self:ScheduleTimer(callback, delay, arg);
			
		activeTimers[id] = 1;
	end
end

function AbsorbsMonitor:FinishUniqueTimer(id)
	activeTimers[id] = nil;
end

local lastAP, lastSP = 0, 0;

function AbsorbsMonitor:Comm_UnitStats()
	self:FinishUniqueTimer("comm_stats");

	if(curChatChannel) then
		local baseAP, plusAP, minusAP = UnitAttackPower("player");
		
		local curAP, curSP = (baseAP + plusAP - minusAP), GetSpellBonusHealing();		
		
		if( (curAP ~= lastAP) or (curSP ~= lastSP) ) then
			self:SendCommMessage("Absorbs_UnitStats", self:Serialize(playerGUID, playerClass, curAP, curSP), curChatChannel);
			
			lastAP, lastSP = curAP, curSP;
			
			CommStatsCooldown = true;
			self:ScheduleTimer(ClearCommStatsCooldown, 15);
		end				
	end
end

function AbsorbsMonitor:Comm_Scaling()
	self:FinishUniqueTimer("comm_scaling");

	if(curChatChannel) then
		self:SendMessage("Absorbs_Scaling", self:Serialize(playerGUID, self:OnScalingBroadcast()), curChatChannel);
		
		CommScalingCooldown = true;
		self:ScheduleTimer(ClearCommScalingCooldown, 15);
	end
end

function AbsorbsMonitor:Unit_UpdatedTotal(guid, value, quality)
	self:SendMessage("Absorbs_Unit_UpdatedTotal", guid, value, quality);
end

function AbsorbsMonitor:Unit_Cleared(guid)
	self:SendMessage("Absorbs_Unit_Cleared", guid);
end

function AbsorbsMonitor:PrintActiveEffects()
	for guid, effects in pairs(activeEffectsBySpell) do
		self:Print("### GUID: "..guid);
		
		self:Print("# total: "..effects[0].." (quality: "..effects[1]..")");
		
		for spellId, effect in pairs(effects) do
			if(spellId > 1) then
				self:Print(spellId..": "..effect[3].."/"..effect[4]);
			end
		end
	end
end

function AbsorbsMonitor:PrintKnownEffects()
	local i = 0;

	for spellId, effect in pairs(AbsorbEffects) do
		self:Print(spellId);
		i = i + 1;
	end
	
	self:Print("### "..i.." effects in total ###");
end

function AbsorbsMonitor:PrintProfiling(criticalOnly)
	if(GetCVar("scriptProfile") ~= "1") then
		self:Print("CPU profiling disabled");
		return;
	end

	UpdateAddOnCPUUsage();
	
	local funcTable;
	
	if(criticalOnly) then
		funcTable = {
			["ApplyActiveEffect"] = AbsorbsMonitor.ApplyActiveEffect,
			["HitUnit"] = AbsorbsMonitor.HitUnit,
			["OnCombatLogEvent"] = AbsorbsMonitor.OnCombatLogEvent,
		};
	else
		funcTable = self;
	end	
		
	local v_type;
	local time_self, time_combined, count;
	
	for k, v in pairs(funcTable) do
		v_type = type(v);
		
		if(v_type == "function") then
			time_self, count = GetFunctionCPUUsage(v, false);
			time_combined = GetFunctionCPUUsage(v, true);
			
			self:Print(k.." (#"..count.."): "..string.format("%.5f", time_self).." / "..string.format("%.5f", time_combined));
		end
	end
end

--[[
local t = {
	['a'] = {1, 2, 3},
	['b'] = {2, 3, 4},
	['c'] = {3, 4, 5},
	['d'] = {4, 5, 6},
	['e'] = {5, 6, 7},
	['f'] = {6, 7, 8},
	['g'] = {7, 8, 9},
	['h'] = {8, 9, 10},
	['i'] = {9, 10, 11},
	['j'] = {10, 11, 12}
};

-- Benchmark for hash access
function AbsorbsMonitor:TestA()
	local d = 0;

	for i = 0, 1000 do
		d = d + t.a[2] + t.b[1] + t.c[3] + t.d[1] + t.e[2] + t.f[3] + t.g[1] + t.h[2] + t.i[1] + t.j[3];
	end
end

-- Benchmark for string concentation
function AbsorbsMonitor:TestB()
	local t = {"a", "b", "c", "d", "e", "f", "g", "h", "i", "j", "k", "l", "m", "n", "o", "p", "q", "r", "s", "t", "u", "v", "w", "x", "y", "z"};
	local m = {};
	
	local k = 1;
	
	for i = 1, 26 do
		for j = 1, 26 do
			--m[i.."-"..j] = t[i].."-"..t[j];
			m[k] = t[i].."-"..t[j].."*"..t[i].."/"..t[j];
			k = k + 1;
		end
	end
end
]]
